import os
from ase.geometry import get_layers
from ase.constraints import FixAtoms
from ase.io.espresso import construct_namelist
from xespresso import Espresso
from xespresso.xio import build_atomic_species_str
from ase.dft.bandgap import bandgap
import pickle
import multiprocessing
import numpy as np
#====================================================
def get_nbnd(atoms = None, scale = 1.2, pseudopotentials = {}, nspin = 1, input_data = {}):
    input_parameters = construct_namelist(input_data)
    atomic_species_str, species_info, total_valence = build_atomic_species_str(atoms, input_parameters, pseudopotentials)
    nbnd = int(total_valence/(2.0/nspin))
    nbnd_scale = int(nbnd*scale)
    print(' total valence: %s\n nbnd: %s\n scaled nbnd: %s'%(total_valence, nbnd, nbnd_scale))
    return nbnd
def merge_slab(slab1, slab2, index = 2):	
    '''	
    '''	
    slab2.cell[index] = slab1.cell[index]	
    slab2.set_cell(slab1.cell, scale_atoms = True)	
    slab1 = slab1 + slab2
    slab1.cell[2][2] = max(slab1.positions[:, 2] + 15)
    slab1.wrap()
    return slab1
def qeinp(calculation, ecutwfc = 30, mixing_beta = 0.5, conv_thr = 1.0e-8,
    edir = False, input_ntyp = {}, atoms = None):
    #
    inp = {
    #control
    'calculation':         calculation,
    'max_seconds':         78000,
    'verbosity':           'high',
    'tprnfor':             True,
    #system
    'ecutwfc':             ecutwfc,
    'ecutrho':             ecutwfc*8,
    'occupations':         'smearing',
    'degauss':             0.01,
    'input_ntyp':          input_ntyp,
    #electrons
    'mixing_beta':         mixing_beta,
    'conv_thr':            conv_thr,
    'electron_maxstep':    400,
    }
    #
    if edir:
        inp.update(dipole_correction(atoms, edir))
        
    return inp
def dipole_correction(atoms, edir = 3):
    inp = {
          'dipfield':  True, 
          'tefield':   True,
          'edir':      edir,
          'eamp':      0.001,
          'eopreg':    0.05,
          }
    emaxpos = max(atoms.positions[:, edir - 1] + atoms.cell[edir - 1][edir - 1])/2.0/atoms.cell[edir - 1][edir - 1]
    inp['emaxpos'] = round(emaxpos, 2)
    return inp

# tools


def build_oer(atoms):
    '''
    '''
    from ase.atoms import Atoms
    #----------------------------------
    ooh = Atoms('O2H', positions = [[0, 0, 0], [1.4, 0, 0], [1.4, 0, 1.0]])
    ooh.rotate('z', np.pi/4)
    mols = {
    'o' : Atoms('O'),
    'oh' : Atoms('OH', positions = [[0, 0, 0], [0, 0, 1.0]]),
    'ooh' : ooh,
    }
    #
    jobs = {}
    maxz = max(atoms.positions[:, 2])
    indm = [atom.index for atom in atoms if atom.z > maxz - 1.0 and atom.symbol != 'O'][0]
    for job, mol in mols.items():
        # print(job, mol)
        ads = mol.copy()
        natoms = atoms.copy()
        ads.translate(atoms[indm].position - ads[0].position + [0, 0, 1.9])
        natoms = natoms + ads
        jobs[job] = natoms
    return jobs

def fix_layers(atoms, miller = (0, 0, 1), tol = 1.0, n = [0, 4]):
    '''
    '''
    layers = get_layers(atoms, miller, tol)[0]
    index = [j for j in range(len(atoms)) if layers[j] in range(n[0], n[1])]
    constraint = FixAtoms(indices=index)
    atoms.set_constraint(constraint)
    return atoms

def mypool(jobs, func):
    '''
    '''
    from random import random
    from time import sleep
    pool = multiprocessing.Pool(len(jobs))
    results = []
    images = []
    t = 0
    for job, atoms in jobs.items():
        print(job, len(atoms), atoms)
        sleep(random()*2)
        results.append(pool.apply_async(func, (job, atoms, t)))
        t += random()*5
    for r in results:
        r.get()
    pool.close()
    pool.join()

def dwubelix(updates = []):
    file = 'pw err out dos pdos projwfc int xyz path a.xml'
    print('Downloading.....')
    cwd = os.getcwd()
    for update in updates:
        os.chdir(update)
        os.system('dwubelix.py %s'%file)
        os.chdir(cwd)
    print('Finished')
def ana(dire, calc):
    atoms = calc.results['atoms']
    results = [dire, atoms, atoms.cell, atoms.positions]
    for prop in ['energy', 'forces', 'stress', 'magmoms']:
        if prop in calc.results:
            prop = calc.results[prop]
        else:
            prop = None
        results.append(prop)
    return results
# tools
def summary(updates = [], prefix = 'datas'):
    import pandas as pd
    columns = ['label', 'atoms', 'cell', 'positions', 'energy', 'forces', 'stress']
    file = '%s.pickle' % prefix
    db = '%s.db' % prefix
    if os.path.exists(file):
        with open(file, 'rb') as f:
            datas, df = pickle.load(f)
    else:
        datas = {}
        df = pd.DataFrame(columns=columns)
    calc = Espresso()
    print('Reading.....')
    for update in updates:
        cwd = os.getcwd()
        for i,j,y in os.walk(update):
            output = is_espresso(i)
            if output:
                os.chdir(i)
                print('dire:', i)
                calc.directory = cwd + '/' + i
                calc.prefix = output[0:-4]
                try:
                    calc.results = {}
                    calc.read_results()
                    # gap, p1, p2 = bandgap(calc)
                    # calc.results['gap'] = gap
                    # t = calc.get_time()
                    # calc.results['time'] = t
                    datas[i] = calc.results
                    # results = ana(i, calc)
                    # df.loc[len(df)] = results
                except Exception as e:
                    print('='*30, '\n', i, e)
            os.chdir(cwd)
    with open(file, 'wb') as f:
        pickle.dump([datas, df], f)
    print('Finished')

def is_espresso(path):
    '''
    check espresso 
    '''
    dirs = os.listdir(path)
    # print(dirs)
    # flag = True
    
    for qefile in ['.pwi']:
        flag = False
        for file in dirs:
            if qefile in file:
                return file
        if not flag:
            return False
    # return flag
def grep_valence_configuration(pseudopotential):
    """
    Given a UPF pseudopotential file, find the valence configuration.
    
    Valence configuration:
    nl pn  l   occ       Rcut    Rcut US       E pseu
    3S  1  0  2.00      0.700      1.200    -6.910117

    """
    orbitals = ['S', 'P', 'D', 'F']
    valence = {}
    with open(pseudopotential) as psfile:
        lines = psfile.readlines()
        for i in range(len(lines)):
            if 'valence configuration:' in lines[i].lower():
                j = i + 2
                ob = lines[j].split()[0]
                while ob[1] in orbitals:
                    valence[ob] = lines[j].split()[3]
                    j += 1
                    ob = lines[j].split()[0]
                return valence
    if not valence:
        raise ValueError('Valence configuration missing in {}'.format(pseudopotential))




"""Quantum ESPRESSO Calculator

export PYTHONPATH="Your-Location":$PYTHONPATH
export ASE_ESPRESSO_COMMAND="/path/to/PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo"
export ESPRESSO_PSEUDO="/path/to/pseudo"

Run PACKAGE.x jobs.
"""


from ase import io
from ase.calculators.calculator import FileIOCalculator, CalculationFailed, equal, compare_atoms
import ase.calculators.espresso
from xespresso.xio import write_espresso_in, read_espresso_input, read_espresso_asei, write_espresso_asei, get_atomic_species
import os
import shutil
import numpy as np
from datetime import datetime
import copy
config_files = [os.path.join(os.environ['HOME'], '.xespressorc'),
            '.xespressorc']



class Espresso(ase.calculators.espresso.Espresso):
    implemented_properties = ['energy', 'forces', 'stress', 'magmoms', 'time']
    command = 'PACKAGE.x  PARALLEL  -in  PREFIX.PACKAGEi  >  PREFIX.PACKAGEo'

    def __init__(self, restart=None, ignore_bad_restart_file=False,
                 label='xespresso', prefix = None, atoms=None, package = 'pw', parallel = '',
                 queue = None, **kwargs):
        """
        Accepts all the options for pw.x as given in the QE docs, plus some
        additional options:

        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE
        queue: dict
            A dictionary with parameters for job submission, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
        package: str
            Choose the quantum espresso pacakge: pw, dos, projwfc, band, pp, ph, ..
            For NEB calculation, please use neb.NEBEspresso module.
            Calculaiton use phonon is not implemented yet.
        parallel: str
            A str which control the parallelization parameters: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).

        General working procedure is as follows:
            1. Perform a regular self-consistent calculation:
              >>> calc = Espresso(input_data=input_data, ...)
              >>> atoms.set_calculator(calc)
              >>> atoms.get_potential_energy()
            2. post calculation can be made as follows:
              >>> calc.post('pacakge' = 'dos', **kwargs)
              >>> calc.post('pacakge' = 'pp', **kwargs)
            3. non-self-consistent calculation can be made as follows:
              >>> calc.nscf(calculation = 'nscf', kpts=(4, 4, 4))
              >>> calc.nscf_calculate()
        """
        self.set_label(restart, label, prefix)
        kwargs = self.set_kwargs(kwargs)
        ase.calculators.espresso.Espresso.__init__(self, restart = self.directory, ignore_bad_restart_file = True,
                                  label = self.label, atoms = atoms, **kwargs)
        if atoms:
            self.atoms = atoms
        self.set_queue(package = package, parallel = parallel, queue = queue)
        self.parallel = parallel
        
    def set_label(self, restart, label, prefix):
        '''
        '''
        if restart:
            self.directory = os.path.split(label)[0]
            self.prefix = os.path.split(label)[1]
        else:
            self.directory = label
            if not prefix:
                self.prefix = os.path.split(label)[1]
            else:
                self.prefix = prefix
        self.label = os.path.join(self.directory, self.prefix)
        self.pwi = os.path.join(self.directory, '%s.pwi'%self.prefix)
        self.pwo = os.path.join(self.directory, '%s.pwo'%self.prefix)
        self.asei = os.path.join(self.directory, '%s.asei'%self.prefix)
        self.asei_temp = os.path.join(self.directory, '.%s.asei_temp'%self.prefix)
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        self.scf_directory = None
        self.scf_parameters = None
        self.scf_results = None
    def set_kwargs(self, kwargs):
        
        if 'input_data' not in kwargs:
            kwargs['input_data'] = {}
        ase_parameters = copy.deepcopy(kwargs)
        for key, value in kwargs.items():
            if key not in ['pseudopotentials', 'kpts', 'kspacing', 'koffset', 'input_data', 'climbing_images', 'path_data']:
                ase_parameters['input_data'][key] = value
                del ase_parameters[key]
        ase_parameters['input_data']['prefix'] = self.prefix
        ase_parameters['input_data']['verbosity'] = 'high'
        self.ase_parameters = ase_parameters
        return ase_parameters
    def set_queue(self, package = 'pw', parallel = '', queue = None):
        '''
        If queue, change command to "sbatch .job_file".
        The queue information are written into '.job_file'
        '''
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        command = os.environ.get('ASE_ESPRESSO_COMMAND')
        if 'PACKAGE' in command:
            if 'pw' in package:
                command = command.replace('PACKAGE', package, 1)
                command = command.replace('PACKAGE', 'pw', 2)
            else:
                command = command.replace('PACKAGE', package)
        if 'PREFIX' in command:
            command = command.replace('PREFIX', self.prefix)
        if 'PARALLEL' in command:
            command = command.replace('PARALLEL', parallel)
        if queue:
            script = ''
            for cf in config_files:
                if os.path.exists(cf):
                    file = open(cf, 'r')
                    script = file.read()
            jobname = self.prefix
            with open('%s/.job_file' % self.directory, 'w') as fh:
                fh.writelines("#!/bin/bash\n")
                fh.writelines("#SBATCH --job-name=%s \n" % jobname)
                fh.writelines("#SBATCH --output=%s.out\n" % self.prefix)
                fh.writelines("#SBATCH --error=%s.err\n" % self.prefix)
                fh.writelines("#SBATCH --wait\n")
                for key, value in queue.items():
                    if value:
                        fh.writelines("#SBATCH --%s=%s\n" %(key, value))
                fh.writelines("%s \n"%script)
                fh.writelines("%s \n" % command)
            self.command = "sbatch {0}".format('.job_file')
        else:
            self.command = command
    def read(self, label):
        """Read the files in a calculation if they exist.
        This function reads self.parameters and atoms.
        """
        # Else read a regular calculation. we start with reading stuff
        # that is independent of the calculation state.
        self.directory = label
        try:
            atoms, parameters = read_espresso_asei(self.asei)
            self.restart_atoms = atoms
            self.restart_parameters = parameters
            self.read_results()
        except:
            pass
    def set(self, **kwargs):
        changed_parameters = FileIOCalculator.set(self, **kwargs)
        
    def check_state(self, atoms, tol=1e-4):
        """Check for any system changes since last calculation."""
        if not os.path.exists(self.asei):
            return True
        if isinstance(self.restart_atoms, list):
            if len(self.restart_atoms) != len(atoms):
                return True
            for i in range(len(atoms)):
                if compare_atoms(self.restart_atoms[i], atoms[i], tol=tol,
                             excluded_properties=set(self.ignored_changes)):
                    return True
        if self.restart_parameters != self.ase_parameters:
            return True
        print('Same geometry and parameters, try to check the results are available or not.')
        return False

    def read_results(self):
        '''
        get atomic species
        '''
        pwo = self.prefix + '.pwo'
        pwos = [file for file in os.listdir(self.directory) if pwo in file]
        output = None
        for pwo in pwos:
            atomic_species = None
            pwo = os.path.join(self.directory, pwo)
            atomic_species = get_atomic_species(pwo)
            try:
                output = io.read(pwo)
                atomic_species = get_atomic_species(pwo)
                if atomic_species: output.info['species'] = atomic_species
            except Exception as e:
                print(pwo, e)
        if output:
            self.calc = output.calc
            self.results = output.calc.results
            self.results['atoms'] = output
            self.efermi = self.get_fermi_level()
            self.nspins = self.get_number_of_spins()
        else:
            print('\nRead result failed\n')
    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        write_espresso_in(self.label + '.pwi', atoms, **self.parameters)
        write_espresso_asei(self.label + '.asei', atoms, self.parameters)

    def read_xml_file(self):
        '''
        If data-file-schema.xml is readable, then fromfile = True
        '''
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        if not os.path.exists(self.save_directory): return False
        xml0 = os.path.join(self.save_directory, 'data-file-schema.xml')
        xmls = [file for file in os.listdir(self.save_directory) if 'data-file-schema.xml' in file]
        goodxml = []
        for xml in xmls:
            xml = os.path.join(self.save_directory, xml)
            # print('read results from: %s'%xml)
            with open(xml, 'r') as f:
                lines = f.readlines()
                if len(lines) < 1: continue
                if '</qes:espresso>' in lines[-1]:
                    goodxml.append(xml)
        fromfile = False
        nxml = len(goodxml)
        # print(nxml, goodxml, xml0)
        if nxml > 0:
            if xml0 not in goodxml:
                shutil.copy(goodxml[-1], xml0)
            fromfile = True
        return fromfile
    def read_convergence(self):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        '-1': 'Cancelled',
        '1': 'Need restart'
        '2': 'Pending or not submit',
        '3': 'Not start',
        '4': 'Not finish',
        }
        '''
        fromfile = self.read_xml_file()
        errfile = self.label + '.err'
        if not os.path.exists(errfile):
            # print('%s not exists'%errfile)
            return 1, fromfile, 'Pending or not submit'
        errs = ['RESTART', 'pw.x', 'out-of-memory', 
                'NODE FAILURE', 'TIME LIMIT', 'COMMUNICATION', 
                'segmentation fault', 'PARSE_ERR', 'mpirun']
        cancelled = False
        with open(errfile, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'CANCELLED' in line:
                    cancelled = True
                for err in errs:
                    if err in line:
                        return 1, fromfile, line
            # cancelled by owner
            if len(lines) > 0:
                if cancelled:
                    return -1, fromfile, lines[0]
        # 
        output = self.label + '.pwo'
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return 3, fromfile, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0: return 1, fromfile, 'pwo file has nothing'
            stime = lines[1].split('starts on')[1]
            nlines = len(lines)
            n = min([100, nlines])
            for line in lines[-n:-1]:
                if line.rfind('too many bands are not converged') > -1:
                    return 1, fromfile, '%s, %s'%(stime, line)
                if line.rfind('convergence NOT achieved after') > -1:
                    fromfile = True
                    return 1, fromfile, '%s, %s'%(stime, line)
                if line.rfind('Maximum CPU time exceeded') > -1:
                    fromfile = True
                    return 1, fromfile, '%s, %s'%(stime, line)
                if line.rfind('JOB DONE.') > -1:
                    fromfile = True
                    return 0, fromfile, line
        return 4, fromfile, line
    def run(self, atoms = None, restart = False):
        '''
        run and restart
        '''
        if not restart:
            try:
                self.calculate()
            except Exception as e:
                print('Not converge: %s'%e)
        converged, fromfile, meg0 = self.read_convergence()
        print(converged, fromfile, meg0)
        i = 0
        # restart 0, 1, 2
        while converged > 0 or restart == 2:
            restart = 1
            # backup old files, *pwo, *xml
            self.backup_file('%s.pwo'%self.prefix, directory = self.directory)
            self.backup_file('data-file-schema.xml', directory = self.save_directory)
            print('RESTART %s'%i)
            if fromfile:
                self.parameters['input_data']['restart_mode'] = 'restart'
                # self.parameters['input_data']['startingwfc'] = 'file'
            else:                
                self.parameters['input_data']['restart_mode'] = 'from_scratch'
            try:
                atoms.get_potential_energy()
            except Exception as e:
                print('Not converge: %s'%e)
            converged, fromfile, meg = self.read_convergence()
            print(converged, fromfile, meg)
            if meg == meg0:
                print('Same message! \n %s'%meg)
                break
            meg0 = meg
            i += 1
    def backup_file(self, src, directory = '.'):
        '''
        compare files
        backup
        '''
        import filecmp
        tnow = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        new_src = os.path.join(directory, '%s-%s'%(tnow, src))
        src = os.path.join(directory, src)
        if not os.path.exists(src): return 0
        flist = []
        ftype = src.split('.')[-1]
        flag = True
        for dst in os.listdir(directory):
            dst = os.path.join(directory, dst)
            if dst == src: continue
            if dst.endswith(ftype):
                if filecmp.cmp(dst, src):
                    flag = False
        # print('backup files: ', flag, src, new_src)
        if flag:
            shutil.copy(src, new_src)
        return 0
    def nscf(self, calculation = 'nscf', package = 'pw', queue = False, parallel = '', kpts = (10, 10, 10), **kwargs):
        import copy
        # save scf parameters
        if not self.scf_directory:
            print(self.directory)
            self.scf_directory = self.directory
        if not self.scf_parameters:
            self.scf_parameters = copy.deepcopy(self.parameters)
        if not self.scf_results:
            self.scf_results = copy.deepcopy(self.results)
            self.efermi = self.get_fermi_level()
        # read new atoms from results and set magmoms
        self.atoms = self.results['atoms']
        self.parameters = copy.deepcopy(self.scf_parameters)
        self.parameters['kpts'] = kpts
        self.parameters['calculation'] = calculation
        self.parameters['verbosity'] = 'high'
        self.parameters['restart_mode'] = 'restart'
        for key, value in kwargs.items():
            self.parameters[key] = value
        # create working directory
        self.save_directory_old = os.path.join(self.scf_directory, '%s.save'%self.prefix)
        self.directory = os.path.join(self.scf_directory, '%s/'%calculation)
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        self.label = os.path.join(self.directory, self.prefix)
        self.set_queue(package = package, parallel = parallel, queue = queue)
    def nscf_calculate(self, ):
        import shutil
        if not os.path.exists(self.save_directory):
            os.makedirs(self.save_directory)
        files = ['charge-density.hdf5', 'charge-density.dat', 'data-file-schema.xml', 'paw.txt', 'occup.txt']
        for species, pseudo in self.parameters['pseudopotentials'].items():
            files.append(pseudo)
        for file in files:
            file = os.path.join(self.save_directory_old, '%s'%file)
            if os.path.exists(file):
                # print(file)
                shutil.copy(file, self.save_directory)
        print('-'*30)
        print('\n {0} calculation in {1} ......'.format(
                    self.parameters['calculation'], 
                    self.directory))
        self.calculate()
        # self.read_results()

    def get_time(self, ):
        t = 0
        filename = os.path.join(self.directory, '%s.pwo'%self.prefix)
        with open(filename) as f:
            lines = f.readlines()
            for line in lines[-200:]:
                if 'init_run     :' in line or 'electrons    :' in line:
                    t += float(line.split('CPU')[1].split('s WALL')[0])
        self.results['time'] = t
        return t
    def post(self, queue = False, package = 'dos', parallel = '', **kwargs):
        '''
        '''
        import subprocess
        post_parameters = {
        'dos':  {'DOS': ['prefix', 'outdir', 'bz_sum', 'ngauss', 'degauss', 
                        'Emin', 'Emax', 'DeltaE', 'fildo', ]},
        'pp':   {'INPUTPP': ['prefix', 'outdir', 'filplot', 'plot_num', 
                       'spin_component', 'spin_component', 'emin', 'emax', 
                       'delta_e', 'degauss_ldos', 'sample_bias', 'kpoint', 
                       'kband', 'lsign', 'spin_component', 'emin', 'emax', 
                       'spin_component', 'spin_component', 'spin_component', 
                       'spin_component'], 
                'PLOT': ['nfile', 'filepp', 'weight', 'iflag', 'output_format', 
                    'fileout', 'interpolation', 'e1', 'x0', 'nx', 'e1', 'e2', 
                    'x0', 'nx', 'ny', 'e1', 'e2', 'e3', 'x0', 'nx', 'ny', 'nz', 
                    'radius', 'nx', 'ny']},
        'projwfc': {'PROJWFC': ['prefix', 'outdir', 'ngauss', 'degauss', 
                    'Emin', 'Emax', 'DeltaE', 'lsym', 'pawproj', 'filpdos', 'filproj', 
                    'lwrite_overlaps', 'lbinary_data', 'kresolveddos', 'tdosinboxes', 
                    'n_proj_boxes', 'irmin(3,n_proj_boxes)', 'irmax(3,n_proj_boxes)', 
                    'plotboxes', ]},
        'hp': {'INPUTHP': ['prefix','outdir','iverbosity','max_seconds','nq1','nq2',
                           'nq3','skip_equivalence_q','determine_num_pert_only',
                           'find_atpert','docc_thr','skip_type','equiv_type',
                           'perturb_only_atom','start_q','last_q','sum_pertq',
                           'compute_hp','conv_thr_chi','thresh_init','ethr_nscf',
                           'niter_max','alpha_mix(i)','nmix','num_neigh','lmin','rmax',]},
        'bands': {'BANDS': ['prefix', 'outdir', 'filband', 'spin_component', 'lsigma', 
                            'lp', 'filp', 'lsym', 'no_overlap', 'plot_2d', 'firstk', 'lastk']},
        'ph': {'INPUTHP': ['amass', 'outdir', 'prefix', 'niter_ph', 'tr2_ph', 
                           'alpha_mix(niter)', 'nmix_ph', 'verbosity', 'reduce_io', 
                           'max_seconds', 'fildyn', 'fildrho', 'fildvscf', 'epsil', 
                           'lrpa', 'lnoloc', 'trans', 'lraman', 'eth_rps', 'eth_ns', 
                           'dek', 'recover', 'low_directory_check', 'only_init', 'qplot', 
                           'q2d', 'q_in_band_form', 'electron_phonon', 'el_ph_nsigma', 
                           'el_ph_sigma', 'ahc_dir', 'ahc_nbnd', 'ahc_nbndskip', 
                           'skip_upperfan', 'lshift_q', 'zeu', 'zue', 'elop', 'fpol', 
                           'ldisp', 'nogg', 'asr', 'ldiag', 'lqdir', 'search_sym', 
                           'nq1', 'nq2', 'nq3', 'nk1', 'nk2', 'nk3', 'k1', 'k2', 'k3', 
                           'diagonalization', 'read_dns_bare', 'ldvscf_interpolate', 
                           'wpot_dir', 'do_long_range', 'do_charge_neutral', 'start_irr', 
                           'last_irr', 'nat_todo', 'modenum', 'start_q', 'last_q', 
                           'dvscf_star', 'drho_star', ]},
        
        }
        kwargs['prefix'] = self.prefix
        package_parameters = post_parameters[package]
        filename = os.path.join(self.directory, '%s.%si' %(self.prefix, package))
        with open(filename, 'w') as f:
            for section, parameters in package_parameters.items():
                f.write('&%s\n '%section)
                for key, value in kwargs.items():
                    if key in parameters:
                        f.write('  %s = %s, \n' %(key, value))
                f.write('/ \n')
        self.set_queue(package = package, parallel = parallel, queue = queue)
        command = self.command
        print('Running %s'%package)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Calculator "{}" failed with command "{}" failed in '
                   '{} with error code {}'.format(self.name, command,
                                                  path, errorcode))
            raise CalculationFailed(msg)
        print('Done: %s' % package)
    def get_work_function(self, ax = None, inpfile = 'potential.cube', output = None, shift = False):
        import matplotlib.pyplot as plt
        from ase.io.cube import read_cube_data
        from ase.units import create_units
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()
        units = create_units('2006')
        #
        filename = os.path.join(self.directory, inpfile)
        data, atoms = read_cube_data(filename)
        data = data * units['Ry']
        ef = self.get_fermi_level()
        # x, y, z, lp = calc.get_local_potential()
        nx, ny, nz = data.shape
        axy = np.array([np.average(data[:, :, z]) for z in range(nz)])
        # setup the x-axis in realspace
        uc = atoms.get_cell()
        xaxis = np.linspace(0, uc[2][2], nz)
        if shift:
            ax.plot(xaxis, axy - ef)
            ef = 0
        else:
            ax.plot(xaxis, axy)
            ax.plot([min(xaxis), max(xaxis)], [ef, ef], 'k:')
        ax.set_xlabel('Position along z-axis ($\AA$)')
        ax.set_ylabel('Potential (eV)')
        if output:
            plt.savefig('%s'%output)
        # plt.show()
        atoms = self.results['atoms']
        pos = max(atoms.positions[:, 2] + 3)
        ind = (xaxis > pos) & (xaxis < pos + 3)
        wf = np.average(axy[ind]) - ef
        print('min: %s, max: %s'%(pos, pos + 3))
        print('The workfunction is {0:1.2f} eV'.format(wf))
    def get_bader_charge(self, inpfile = None):
        '''
        '''
        if not inpfile:
            inpfile = '%s.cube' %self.prefix
        command = 'bader %s'%inpfile
        print(command)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err
        acf = os.path.join(self.directory, 'ACF.dat')


    def clean(self):
        '''
        remove wfc, hub files
        '''
        keys = ['.wfc', '.hub']
        files = os.listdir(self.directory)
        for file in files:
            for key in keys:
                if key in file:
                    os.remove(os.path.join(self.directory, file))
        files = os.listdir(self.save_directory)
        keys = ['wfc']
        for file in files:
            for key in keys:
                if key in file:
                    os.remove(os.path.join(self.save_directory, file))



"""
Quantum ESPRESSO Calculator for NEB
"""

import numpy as np
from ase import Atoms
from ase import io
from xespresso import Espresso
from xespresso.xio import write_neb_in, write_espresso_asei
from ase.calculators.calculator import FileIOCalculator, compare_atoms


    
class NEBEspresso(Espresso):
    """
    """
    implemented_properties = ['energy', 'forces', 'stress', 'magmoms', 'time', 'neb']
    command = 'neb.x  PARALLEL  -in  PREFIX.nebi  >  PREFIX.nebo'

    def __init__(self, restart=None, ignore_bad_restart_file=False,
                 label='xespresso', prefix = None, images=[Atoms('')], package = 'neb', parallel = '',
                 queue = None,
                 **kwargs):
        """

        """
        atoms = images[0]
        Espresso.__init__(self, restart, ignore_bad_restart_file,
                 label = label, prefix = prefix, atoms = atoms, package = package, parallel = parallel,
                 queue = queue, **kwargs)
        self.images = images
    def write_input(self, images, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, self.images, properties, system_changes)
        write_neb_in(self.label + '.nebi', self.images, **self.parameters)
        write_espresso_asei(self.label + '.asei', self.images, self.parameters)
    def get_neb_path_energy(self, images = None):
        if images is None:
            images = self.images
        paths, energies = self.get_property('neb', images)
        return paths, energies
    def read_results(self):
        paths, energies, error = self.read_dat()
        try:
            paths, energies, error = self.read_dat()
            self.paths = paths
            self.energies = energies
            self.error = error
            self.results['neb'] = [self.paths, self.energies]
            interpolation_paths, interpolation_energies = self.read_int()
            self.interpolation_paths = interpolation_paths
            self.interpolation_energies = interpolation_energies
        except Exception as e:
            print('\nread paths and energies failed\n')
            print(e)
        try:
            images = self.read_xyz()
            self.images = images
        except Exception as e:
            print('\nread xyz failed\n')
            print(e)
    def read_dat(self):
        data = np.loadtxt(self.directory+'/%s.dat' % self.prefix)
        paths, energies, error = data[:, 0], data[:, 1], data[:, 2]
        return paths, energies, error
    def read_int(self):
        data = np.loadtxt(self.directory+'/%s.int' % self.prefix)
        paths, energies = data[:, 0], data[:, 1]
        return paths, energies
    def read_xyz(self):
        images = io.read(self.directory+'/%s.xyz' % self.prefix)
        return images
    def read_path(self):
        pass
    def plot(self):
        import matplotlib.pyplot as plt
        plt.plot(self.paths, self.energies, 'o')
        plt.plot(self.interpolation_paths, self.interpolation_energies, '--')
        plt.xlabel('Path')
        plt.ylabel('Energy (eV)')
        Eb = max(self.energies) - self.energies[0]
        Ef = max(self.energies) - self.energies[-1]
        plt.title('Eb = %1.2f(eV), Ef = %1.2f'%(Eb, Ef))
        plt.show()


def interpolate(images, n = 10):
    '''
    Interpolate linearly the potisions of the middle temp:
    '''
    from ase.neb import NEB
    if len(images) == 2:
        new_images = [images[0]]
        for i in range(n):
            new_images.append(images[0].copy())
        new_images.append(images[1])
        neb = NEB(new_images)
        neb.interpolate()
        return new_images
    elif len(images) == 3:
        new_images1 = [images[0]]
        for i in range(n):
            new_images.append(images[0].copy())
        new_images.append(images[1])
        neb = NEB(new_images1)
        neb.interpolate()
        new_images = neb.images.copy()
        #
        new_images2 = [images[1]]
        for i in range(n):
            new_images.append(images[1].copy())
        new_images.append(images[2])
        neb = NEB(new_images2)
        neb.interpolate()
        new_images.extend(neb.images)
        return new_images




import os
from os import path
import json
import pickle
import operator as op
import warnings
import numpy as np
from ase.atoms import Atoms
from ase.calculators.calculator import kpts2ndarray, kpts2sizeandoffsets
from ase.constraints import FixAtoms, FixCartesian
from ase.data import chemical_symbols, atomic_numbers
from ase.io.espresso import read_espresso_in, construct_namelist, grep_valence, SSSP_VALENCE, read_fortran_namelist
from ase.units import create_units


def write_espresso_in(filename, atoms, input_data={}, pseudopotentials=None,
                      kspacing=None, kpts=None, koffset=(0, 0, 0),
                      crystal_coordinates=False, **kwargs):
    """
    Create an input file for pw.x.

    Modified the read_espresso_in function from ase.io.espresso
    Parameters:
        Please have a look at the ase.io.espresso module in ASE
    
    Implemented:

    - support parameters with ntyp
    - support atomic species
    """

    # Convert to a namelist to make working with parameters much easier
    # Note that the name ``input_data`` is chosen to prevent clash with
    # ``parameters`` in Calculator objects
    input_parameters = construct_namelist(input_data, **kwargs)

    # Construct input file into this
    pwi = []
    # atomic_species
    atomic_species_str, species_info, total_valence = build_atomic_species_str(atoms, input_parameters, pseudopotentials)
    # sections
    section_str, input_parameters = build_section_str(atoms, species_info, input_data, input_parameters)
    pwi.extend(section_str)
    # Pseudopotentials
    pwi.extend(atomic_species_str)
    # KPOINTS - add a MP grid as required
    pwi.extend(build_kpts_str(atoms, kspacing, kpts, koffset))
    # CELL block, if required
    pwi.extend(build_cell_str(atoms, input_parameters))
    # Positions - already constructed, but must appear after namelist
    engine_str = pwi.copy()
    pwi.extend(build_atomic_positions_str(atoms, crystal_coordinates))
    # write file or not
    if filename:
        with open(filename, 'w') as fd:
            fd.write(''.join(pwi))
    # return section_str, atomic_species_str, kpts_str, cell_str, atomic_positions_str
    return engine_str

def build_section_str(atoms, species_info, input_data, input_parameters):
    '''
    '''
    # Add computed parameters
    # different magnetisms means different types
    input_parameters['system']['ntyp'] = len(species_info)
    input_parameters['system']['nat'] = len(atoms)

    #
    if 'input_ntyp' in input_data:
        for key, value in input_data['input_ntyp'].items():
            for species in value:
                if species in species_info:
                    mag_str = '{0}({1})'.format(key, species_info[species]['index'])
                    input_parameters['system'][mag_str] = value[species]
    # Use cell as given or fit to a specific ibrav
    if 'ibrav' in input_parameters['system']:
        ibrav = input_parameters['system']['ibrav']
        if ibrav != 0:
            celldm = cell_to_ibrav(atoms.cell, ibrav)
            regen_cell = ibrav_to_cell(celldm)[1]
            if not np.allclose(atoms.cell, regen_cell):
                warnings.warn('Input cell does not match requested ibrav'
                              '{} != {}'.format(regen_cell, atoms.cell))
            input_parameters['system'].update(celldm)
    else:
        # Just use standard cell block
        input_parameters['system']['ibrav'] = 0
    # Assume sections are ordered (taken care of in namelist construction)
    # and that repr converts to a QE readable representation (except bools)
    section_str = []
    for section in input_parameters:
        section_str.append('&{0}\n'.format(section.upper()))
        for key, value in input_parameters[section].items():
            if value is True:
                section_str.append('   {0:16} = .true.\n'.format(key))
            elif value is False:
                section_str.append('   {0:16} = .false.\n'.format(key))
            else:
                # repr format to get quotes around strings
                section_str.append('   {0:16} = {1!r:}\n'.format(key, value))
        section_str.append('/\n')  # terminate section
    section_str.append('\n')
    return section_str, input_parameters

def build_atomic_species_str(atoms, input_parameters, pseudopotentials):
    '''
    '''
    # Deal with pseudopotentials
    # Look in all possible locations for the pseudos and try to figure
    # out the number of valence electrons
    pseudo_dirs = []
    if 'pseudo_dir' in input_parameters['control']:
        pseudo_dirs.append(input_parameters['control']['pseudo_dir'])
    if 'ESPRESSO_PSEUDO' in os.environ:
        pseudo_dirs.append(os.environ['ESPRESSO_PSEUDO'])
    pseudo_dirs.append(path.expanduser('~/espresso/pseudo/'))

    # Species info holds the information on the pseudopotential and
    # associated for each element
    if 'species' not in atoms.arrays:
        # print('no species')
        atoms.info['species'] = atoms.get_chemical_symbols()
    if pseudopotentials is None:
        pseudopotentials = {}
    species_info = {}
    atomic_species_str = ['ATOMIC_SPECIES\n']
    ntyp = 0
    # Convert atoms into species.
    for i in range(len(atoms)):
        species = atoms.info['species'][i]
        if species not in species_info:
            ntyp += 1
            species_info[species] = {}
            species_info[species]['index'] = ntyp
            species_info[species]['mass'] = atoms[i].mass
            species_info[species]['element'] = atoms[i].symbol
            species_info[species]['count'] = 1
        else:
            species_info[species]['count'] += 1
    total_valence = 0
    for species in species_info:
        pseudo = pseudopotentials.get(species, '{}_dummy.UPF'.format(species))
        for pseudo_dir in pseudo_dirs:
            if path.exists(path.join(pseudo_dir, pseudo)):
                valence = grep_valence(path.join(pseudo_dir, pseudo))
                break
        else:  # not found in a file
            valence = SSSP_VALENCE[atomic_numbers[species_info[species]['element']]] 
        species_info[species]['pseudo'] = pseudo,
        species_info[species]['valence'] = valence
        total_valence += valence*species_info[species]['count']
        atomic_species_str.append(
            '{species} {mass} {pseudo}\n'.format(
                species=species, mass=species_info[species]['mass'],
                pseudo=pseudo))
    atomic_species_str.append('\n')
    return atomic_species_str, species_info, total_valence
              

def build_cell_str(atoms, input_parameters):
    '''
    '''
    # CELL block, if required
    if input_parameters['SYSTEM']['ibrav'] == 0:
        cell_str = ['CELL_PARAMETERS angstrom\n']
        cell_str.append('{cell[0][0]:.14f} {cell[0][1]:.14f} {cell[0][2]:.14f}\n'
                   '{cell[1][0]:.14f} {cell[1][1]:.14f} {cell[1][2]:.14f}\n'
                   '{cell[2][0]:.14f} {cell[2][1]:.14f} {cell[2][2]:.14f}\n'
                   ''.format(cell=atoms.cell))
        cell_str.append('\n')
    return cell_str
def build_kpts_str(atoms, kspacing, kpts, koffset):
    '''
    '''
    # KPOINTS - add a MP grid as required
    if kspacing is not None:
        kgrid = kspacing_to_grid(atoms, kspacing)
    elif kpts is not None:
        if isinstance(kpts, dict) and 'path' not in kpts:
            kgrid, shift = kpts2sizeandoffsets(atoms=atoms, **kpts)
            koffset = []
            for i, x in enumerate(shift):
                assert x == 0 or abs(x * kgrid[i] - 0.5) < 1e-14
                koffset.append(0 if x == 0 else 1)
        else:
            kgrid = kpts
    else:
        kgrid = "gamma"

    # True and False work here and will get converted by ':d' format
    if isinstance(koffset, int):
        koffset = (koffset, ) * 3

    # BandPath object or bandpath-as-dictionary:
    if isinstance(kgrid, dict) or hasattr(kgrid, 'kpts'):
        kpts_str = ['K_POINTS crystal_b\n']
        assert hasattr(kgrid, 'path') or 'path' in kgrid
        kgrid = kpts2ndarray(kgrid, atoms=atoms)
        kpts_str.append('%s\n' % len(kgrid))
        for k in kgrid:
            kpts_str.append('{k[0]:.14f} {k[1]:.14f} {k[2]:.14f} 0\n'.format(k=k))
        kpts_str.append('\n')
    elif isinstance(kgrid, str) and (kgrid == "gamma"):
        kpts_str = ['K_POINTS gamma\n']
        kpts_str.append('\n')
    else:
        kpts_str = ['K_POINTS automatic\n']
        kpts_str.append('{0[0]} {0[1]} {0[2]}  {1[0]:d} {1[1]:d} {1[2]:d}\n'
                   ''.format(kgrid, koffset))
        kpts_str.append('\n')
    return kpts_str

def build_atomic_positions_str(atoms, crystal_coordinates):
    '''

    '''
    if 'species' not in atoms.arrays:
        atoms.info['species'] = atoms.get_chemical_symbols()
    # Convert ase constraints to QE constraints
    # Nx3 array of force multipliers matches what QE uses
    # Do this early so it is available when constructing the atoms card
    constraint_mask = np.ones((len(atoms), 3), dtype='int')
    for constraint in atoms.constraints:
        if isinstance(constraint, FixAtoms):
            constraint_mask[constraint.index] = 0
        elif isinstance(constraint, FixCartesian):
            constraint_mask[constraint.a] = constraint.mask
        else:
            warnings.warn('Ignored unknown constraint {}'.format(constraint))
    # Positions - already constructed, but must appear after namelist
    if crystal_coordinates:
        atomic_positions_str = ['ATOMIC_POSITIONS crystal\n']
    else:
        atomic_positions_str = ['ATOMIC_POSITIONS angstrom\n']
    for i in range(len(atoms)):
        atom = atoms[i]
        species = atoms.info['species'][i]
        # only inclued mask if something is fixed
        if not all(constraint_mask[atom.index]):
            mask = ' {mask[0]} {mask[1]} {mask[2]}'.format(
                mask=constraint_mask[atom.index])
        else:
            mask = ''
        if crystal_coordinates:
            coords = [atom.a, atom.b, atom.c]
        else:
            coords = atom.position
        atomic_positions_str.append(
            '{species} '
            '{coords[0]:.10f} {coords[1]:.10f} {coords[2]:.10f} '
            '{mask}\n'.format(species=species, coords=coords, mask=mask))
    return atomic_positions_str

def write_neb_in(filename, images, climbing_images = [], path_data = {}, 
                 input_data={}, pseudopotentials=None,
                 kspacing=None, kpts=None, koffset=(0, 0, 0),
                 crystal_coordinates=False, **kwargs):
    """
    Create an input file for neb.x.

    Parameters
    ----------
    filename: str
        A file to write the input file to.
    images: list
        A list of atomistic configuration including the first, last 
        and intermedinate positions.
    climbing_images: list
        Indices of the images which the Climbing-Image procedure apply
    path_data: dict
        A flat or nested dictionary with path parameters for neb.x
    """
    nebi = ['BEGIN\n']
    package_parameters = {
            'PATH': {'string_method', 'restart_mode', 'nstep_path', 'num_of_images', 
            'opt_scheme', 'CI_scheme', 'first_last_opt', 'minimum_image', 'temp_req', 
            'ds', 'k_max', 'k_min', 'path_thr', 'use_masses', 'use_freezing', 
            'lfcpopt', 'fcp_mu', 'fcp_tot_charge_first', 'fcp_tot_charge_last', },
        }
    path_str = ['\nBEGIN_PATH_INPUT\n']
    for section, parameters in package_parameters.items():
        path_str.append('&%s\n '%section)
        for key, value in path_data.items():
            if key in parameters:
                if value is True:
                    path_str.append('   {0:16} = .true.\n'.format(key))
                elif value is False:
                    path_str.append('   {0:16} = .false.\n'.format(key))
                else:
                    path_str.append('   {0:16} = {1!r:}\n'.format(key, value))
        path_str.append('/ \n')
    if climbing_images:
        path_str.append('CLIMBING_IMAGES\n')
        climbing_images_str = '   '
        for x in climbing_images: climbing_images_str += ' %s'%x
        path_str.append(climbing_images_str)
        path_str.append('\n')
    path_str.append('END_PATH_INPUT\n')
    nebi.extend(path_str)
    
    # BEGIN_ENGINE_INPUT
    engine_str = ['\nBEGIN_ENGINE_INPUT\n']
    pwi = write_espresso_in(None, images[0], input_data, pseudopotentials, 
                            kspacing, kpts, koffset, crystal_coordinates, **kwargs)
    engine_str.extend(pwi)
    
    # POSITION
    atomic_positions_str = ['\nBEGIN_POSITIONS\n']
    for i in range(len(images)):
        atoms = images[i]
        if i == 0:
            atomic_positions_str.append('\nFIRST_IMAGE\n')
        elif i == len(images) - 1:
            atomic_positions_str.append('\nLAST_IMAGE\n')
        else:
            atomic_positions_str.append('\nINTERMEDIATE_IMAGE\n')
        atomic_positions_str.extend(build_atomic_positions_str(atoms, crystal_coordinates))
    atomic_positions_str.append('END_POSITIONS\n')
    #
    engine_str.extend(atomic_positions_str)
    engine_str.append('END_ENGINE_INPUT\n')
    nebi.extend(engine_str)
    nebi.append('END\n')
    with open(filename, 'w') as fd:
        fd.write(''.join(nebi))


def get_atomic_species(pwo):
    '''
    '''
    with open(pwo) as f:
        lines = f.readlines()
        for idx, line in enumerate(lines):
            if 'number of atoms/cell' in line:
                nat = int(line.split()[-1])
            elif 'positions (alat units)' in line:
                atomic_species = [at_line.split()[1]
                    for at_line in lines[idx + 1:idx + 1 + nat]]
                break
    return atomic_species

def read_espresso_input(fileobj, neb = False):
    """Parse a Quantum ESPRESSO input files, '.in', '.pwi'.
    """
    atoms = read_espresso_in(fileobj)
    if isinstance(fileobj, str):
        fileobj = open(fileobj, 'rU')
    # parse namelist section and extract remaining lines
    data, card_lines = read_fortran_namelist(fileobj)
    input_data = {}
    for sec, paras in data.items():
        for key, value in paras.items():
            input_data[key] = value
    # get number of type
    ntyp = data['system']['ntyp']
    pseudopotentials = {}
    trimmed_lines = (line for line in card_lines
                     if line.strip() and not line[0] == '#')
    for line in trimmed_lines:
        if line.strip().startswith('ATOMIC_SPECIES'):
            for i in range(ntyp):
                species = next(trimmed_lines).split()
                pseudopotentials[species[0]] = species[2]
        if line.strip().startswith('K_POINTS'):
            kpts = next(trimmed_lines)
    # calc = Espresso(pseudopotentials=pseudopotentials, 
    #                 input_data = input_data,
    #                 kpts=kpts)
    return atoms, input_data, pseudopotentials, kpts

def read_espresso_asei(fileobj):
    """Parse Quantum ESPRESSO input parameters
    """
    with open(fileobj, 'rb') as fp:
        atoms, parameters = pickle.load(fp)
    return atoms, parameters
def write_espresso_asei(fileobj, atoms, parameters):
    """save Quantum ESPRESSO input parameters
    """
    with open(fileobj, 'wb') as fp:
        pickle.dump([atoms, parameters], fp)
